;************************************************************************
;
;    This procedure reads in and expands the coeffients of the spherical harmonic
;    expansion to calculate potiential values on a regular grid in
;    geomagnetic coordinates
;
;  INPUT:
;    file_name: name of file containing coefficients
;
;    NOTE: the naming convention of the file names is as follows: 
;    - the 'ac' prefix indcates we used IMF data from ACE.
;    - then 'north' or 'south' indicates the hemisphere
;    - the numbers 0t3, 3t5, or 5t10 indicate the IMF magnitude bin in nT
;    - the string Bz+_By+ etc. indicates the IMF clock angle bin
;    - the string DP-,DP0,DP+ indicates negative neutral, or positive
;      tilt (for either hemispere, positive tilt is summer-like
;      conditions)
;
;  OPTIONAL INPUTS
;    del_lat: latitude spacing (in degrees) of grid on which potential
;             is solved (default is 1 degree)
;
;    del_lon: longitude spacing (in degrees) of grid on which potential
;             is solved (default is 15 degrees)
;
;  RETURN VALUE:
;    pos:  A 2D array specifying the locations in altidude adjusted 
;          geomagnetic coordinates (AACGM) at which the potential is calculated
;              1st column: latitude in degrees
;              2nd column: longitude in degrees;
;    
;    pot:  The electrostatic potential in Volts, given at the locations specified
;          in the pos array
;
;  OPTIONAL RETURNED VALUES:
;    latmin:   Low-latitude zero-potential boundary of expansion
; 
;    phi_tot:  This returns the total potential drop across the polar
;              cap in Volts.
;
;    rms_err:  This returns the rms error value of the fit (in m/s)
;
;
;  EXAMPLE CALLING SEQUENCE:
;  model_solver,'ac.north_3t5_By+_DP0.bsph',pos,pot,phi_tot=phi_tot
;*************************************************************************
;

function create_grid,latmin,dlat,dlon
  nlats = fix((90 - latmin)/float(dlat))
  nlons = fix(360/float(dlon))
  pos = fltarr(2, nlons, nlats)

  for i = 0,nlats-1 do begin
     lat = i*dlat + latmin
     pos[0,*,i] = lat
     pos[1,*,i] = findgen(nlons)*dlon
  endfor

  return,reform(pos,2,nlats*nlons)
end

function calc_pot,pos,solution,latmin,order
  ;; Convert position from lat/lon to theta/phi
  theta =  reform(pos[0,*])
  theta =  (90.-theta)*!dtor
  phi   =  reform(pos[1,*])*!dtor

  ;; Normalize theta
  tmax        =  (90.0 - latmin)*!dtor
  alpha       =  !pi/tmax
  theta_prime =  alpha * theta

  x         =  cos(theta_prime)
  plm       =  eval_legendre(order, x)
  pot       =  eval_potential(solution(2,*),plm,phi)

  q         =  where(pos[0,*] le latmin,qc)
  if (qc ne 0) then pot[q] = 0.

return,pot
end

;+
;  FUNCTION EVAL_LEGENDRE
;
;  Purpose:  evalaute all the Associated Legendre Polynomials
;            from L=0 to L=Lmax, at a set of points
;
;  Calling Sequence:
;    PLM = eval_legendre(Lmax, pos)
;
;          where Lmax give the maximum order of the Legendre
;          polynomials, and pos gives the positions where the
;          the polynomials are to be evaluated.
; 
;          pos = pos(2,N) where the first index indicates
;                the theta,phi position and the second index
;                lists the points where we have data.
;
;
;-
; $Log: eval_legendre.pro,v $
; Revision 1.1  1999/08/05 00:15:30  barnes
; Initial revision
;
; Revision 1.1  1997/08/05 14:42:55  baker
; Initial revision
;
;
function dbang,Mmax
result = replicate(1.0d0,Mmax+1)
for i=1,MMax do $
  for j = i, Mmax do result(j) = result(j)*(2*i - 1.0)
return, result
end

function eval_legendre,Lmax,x_in,pm=pm,pp=pp
x = double(reform(x_in,n_elements(x_in)))
N = n_elements(x)
xx = (1-x^2)
xx_L = xx # replicate(1.0d0,Lmax+1)
mover2 = dindgen(Lmax+1)/2.0
mover2 = replicate(1.0d0,N) # mover2
xx_Mover2 = xx_L^mover2
; xx_Mover2 is the matrix of all the (1-x^2) values raised to
; all the powers of m/2 with m running from 0 to Lmax.
;
two_m_dbang = dbang(Lmax)
two_m_dbang = replicate(1.0d0,N) # two_m_dbang
pwrm = replicate(1.0d0,N) # ((-1)^indgen(Lmax+1))
pmm = xx_Mover2*pwrm*two_m_dbang
; 
;  we have now computed the value of Pmm at each point for
;  each value of m from 0 to Lmax
;
;
; We have now compute P(m+1,m) at each point and for each
; value of m from 0 to Lmax.
;
;
;  p(m+1,m) = x(2m+1)P(m,m)
;

pmmp1 = (x # replicate(1.0d0,Lmax+1))*(replicate(1.0d0,N) # $
                                     (dindgen(Lmax+1)*2.0+1.0)) * pmm
;
; OK, now we have pmm and pmmp1 at every point.
; Next we have to compute the rest of the plm values from
; the recursion relation:
; (l-m)P(l,m) = x(2l-1)p(l-1,m) - (l+m-1)P(l-2,m)
;
plm = replicate(0.0d0,N,lmax+1,lmax+1)
for l = 0,lmax do plm(*,l,l) = pmm(*,l)
for l = 0,lmax-1 do plm(*,l+1,l) = pmmp1(*,l)
for l = 0, lmax -2 do $
   for k=l+2,Lmax do plm(*,k,l) = $
               (1.0d0/(k-l))*((x*plm(*,k-1,l)*(2*k -1))-(k+l-1)*plm(*,k-2,l))


if (keyword_set(pm)) then pm = pmm
if (keyword_set(pp)) then pp = pmmp1
return,plm
end

;+
; FUNCTION EVAL_POTENTIAL
;
; PURPOSE:  evaluate the electric potential on a set of
;           points, given the coefficients of the spherical
;           harmonic expansion.
;
; Calling Sequence:
;
;   pot = eval_potential,a,plm,phi
;
;     where "a" is the set of coefficients given as a vector
;               indexed by k = index_legendre(l,m,/dbl)
;
;     plm is an array (N,Lmax,Lmax) where N = number of points
;         where the potential is to be evaluated, and the
;         other two indices give the associated Legendre polynomial
;         to be evaluated at each position.
;
;     phi is the azimuthal coordinate at each evaluation point.
;
; $Log: eval_potential.pro,v $
; Revision 1.1  1999/08/05 00:15:30  barnes
; Initial revision
;
;-
function eval_potential,a,plm,phi

ss=size(plm)
lmax=ss(2)-1
v = replicate(0.,n_elements(phi))

for m=0,lmax do begin
  for L=m,Lmax do begin
     k = index_legendre(L,m,/dbl)
     if (m eq 0) then v = v + a(k)*plm(*,L,0) else $
        v = v + a(k)*cos(m*phi)*plm(*,l,m) + a(k+1)*sin(m*phi)*plm(*,l,m)
  endfor
endfor

return,v
end

function index_legendre,l,m,sh=sh,sngl=sngl,dbl=dbl
  if (m GT l) then return,-1

  if (l eq 0) then return,0 $
  else if (m eq 0) then return, L*L $
  else return, L*L + 2*m - 1
end

Pro model_solver,file_name,pos,pot,del_lat=del_lat,del_lon=del_lon,latmin=latmin, $
   phi_tot=phi_tot,rms_err=rms_err

  if not keyword_set(del_lat) then del_lat = 1
  if not keyword_set(del_lon) then del_lon = 15

  model_reader,file_name,coeffs,latmin,phi_tot=phi_tot,rms_err=rms_err,order=order

  pos = create_grid(latmin,del_lat,del_lon)

  pot = calc_pot(pos,coeffs,latmin,order)

end
